package ac.essex.gp.selection;

import ac.essex.gp.selection.Selectable;
import ac.essex.gp.params.GPParams;
import ac.essex.gp.individuals.Individual;

import java.util.Vector;
import java.util.Hashtable;

/**
 * Base class for a selector, which chooses parents for crossover, mutation
 * and reproduction. The most common one is Tournament Selection, owing to
 * its simplicity and ease of tuning, but there are many others. SXGP uses
 * tournament selection by default, have a look at the RandomTournamentSelector
 * class for more details.
 *
 * @author Olly Oechsle, University of Essex, Date: 22-Jan-2007
 * @version 1.0
 * @see ac.essex.gp.selection.RandomTournamentSelector
 */
public abstract class Selector {

    /**
     * Default niche - select from the whole population - so called
     * panmictic selection.
     */
    public static final int ANY_NICHE = -1;

    /**
     * Niche - determines where the selector will choose individuals
     * from. If it is ANY_NICE, the selector will select from the whole
     * population.
     */
    protected int niche = ANY_NICHE;

    /**
     * A reference to the the (already evaluated) population. This member
     * is kept private because we want to ensure that extending classes get access
     * to the population through the getPopulation() method, which allows
     * this class to ensure that niches are adhered to.
     */
    private Individual[] population;

    /**
     * If operating in population niches, we'll chop up the population
     * into bits so that we can make the selection more efficient. If niches
     * are not used this variable remains null and isn't used.
     */
    protected Hashtable<Integer, Vector<Individual>> nicheCache = null;

    /**
     * Gives the selector a reference to the current population.
     * The population must contain objects of type Selectable
     * @see ac.essex.gp.selection.Selectable
     */
    public void setPopulation(Individual[] population) {
        this.population = population;
    }

    /**
     * Allows the selector to initialise, giving it a reference to the params object,
     * from which it may take any parameters it requires.
     */
    public abstract void initialise(GPParams params);

    /**
     * Selects an individual. This method may be implemented by different
     * selection methods.
     */
    public abstract Selectable select();

    /**
     * Returns the niche that the selector will choose individuals from.
     * If it is -1 the selector will select from the whole population.
     */
    public int getNiche() {
        return niche;
    }

    /**
     * Sets the niche that the selector will choose individuals from.
     * Set it to -1 if you want the selector to select from the whole
     * population (the default) -- so called panmictic selection.
     */
    public void setNiche(int niche) {
        this.niche = niche;
    }

    /**
     * Gets the population (needed by selection implementations). Ensures that what
     * is returned complies with the niche this selector has been instructed to choose.
     */
    protected Individual[] getPopulation() {
        if (niche == ANY_NICHE) {
            // return the whole population
            return population;
        } else {
            // return individuals with the given nicheID
            // TODO: This does not allow travellers between niches.


            // cache the niches so they don't have to be calculated every time.
            // since java still doesn't support associative arrays this is a little
            // messy
            if (nicheCache == null) {

                // implement the cache as a hashtable
                nicheCache = new Hashtable<Integer, Vector<Individual>>();

                // go through the population
                for (int i = 0; i < population.length; i++) {
                    Individual ind = population[i];
                    Vector<Individual> nichePopulation = nicheCache.get(ind.getNicheID());
                    if (nichePopulation == null) {
                        // create a new vector containing only individuals of this type
                        nichePopulation = new Vector<Individual>();
                        nicheCache.put(ind.getNicheID(), nichePopulation);
                    }
                    nichePopulation.add(ind);
                }
            }

            Vector<Individual> inds = nicheCache.get(niche);
            Individual[] pop = new Individual[inds.size()];
            inds.toArray(pop);
            return pop;

        }
    }
}
