package ac.essex.gp.tree;

import ac.essex.gp.individuals.Individual;
import ac.essex.gp.problems.Problem;
import ac.essex.gp.problems.DataStack;
import ac.essex.gp.nodes.ercs.BasicERC;
import ac.essex.gp.Evolve;
import ac.essex.gp.util.FoundBestIndividualException;

import java.util.Vector;

/**
 * <p/>
 * Attempts to improve an individual by fiddling with the values of its ERCs in order to
 * find better solutions.
 * </p>
 *
 * @author Olly Oechsle, University of Essex, Date: 30-Oct-2007
 * @version 1.01 11th December 2007 Fixed a few glitches.
 */
public class ERCOptimiser {

    /**
     * Attempts to improve an individual by fiddling with the values of its ERCs in order to
     * find better solutions.
     * @return True if some improvements could be made, false otherwise.
     */
    public static boolean optimise(Individual ind, Problem p, Evolve e) throws FoundBestIndividualException {

         // get the ercs
        Vector<BasicERC> ercs = TreeUtils.getERCs(ind.getTree(0));

        if (ercs.size() == 0)  {
            //System.err.println("ERC Optimiser: Individual has no ERCs");
            return false;
        }

        int iterationsWithNoImprovement = 0;

        // remember how good the individual used to be.
        p.evaluate(ind, new DataStack(), e);
        double startFitness = ind.getKozaFitness();

        // did we make improvements?
        boolean improved = false;

        int i = 0;

        while (iterationsWithNoImprovement < 10) {

            // choose an erc to change
            BasicERC randomERC = ercs.elementAt(i);

            i++;
            if (i >= ercs.size()) i = 0;

            // remember the old values
            double oldValue = randomERC.getValue();
            double oldFitness = ind.getKozaFitness();

            int oldHits = ind.getHits();
            int oldMistakes =ind.getMistakes();

            // Change it by up to 10% in either positive or negative direction.
            //double newValue = oldValue != 0? oldValue * (1.1 + (Evolve.getRandomNumber() * 0.2)): 0.1 - (Evolve.getRandomNumber() * 0.2);
            randomERC.jitter();
            
            // evaluate the individual
            p.evaluate(ind, new DataStack(), e);

            // find the new fitness
            double newFitness = ind.getKozaFitness();

            // is it better?
            if (newFitness < oldFitness) {

                // reset the counter
                iterationsWithNoImprovement = 0;
                improved = true;

                if (newFitness == 0)  {
                    throw new FoundBestIndividualException(ind);
                }

            } else {

                iterationsWithNoImprovement++;

                // undo the changes
                randomERC.setValue(oldValue);
                ind.setKozaFitness(oldFitness);
                ind.setHits(oldHits);
                ind.setMistakes(oldMistakes);

            }

        } // loop

/*        if (improved) {
            System.err.println("Optimisation complete - Fitness improved from " + startFitness + " to " + ind.getKozaFitness());
        }*/

        return improved;

    }

}
